import dataclasses
import json
import os
from pathlib import Path
import platform
import subprocess
import sys
import tempfile
from typing import Annotated, Optional

from cyclopts import App, Parameter
import cyclopts.config
from platformdirs import user_config_dir

from reprexlite.config import ReprexConfig
from reprexlite.exceptions import (
    EditorError,
    InputSyntaxError,
    IPythonNotFoundError,
)
from reprexlite.formatting import formatter_registry
from reprexlite.reprexes import Reprex
from reprexlite.version import __version__


def get_version():
    return __version__


HELP_TEMPLATE = """
Render reproducible examples of Python code for sharing. Your code will be executed and, in
the default output style, the results will be embedded as comments below their associated
lines.

By default, your system's default command-line text editor will open for you to type or paste
in your code. This editor can be changed by setting either of the `VISUAL` or `EDITOR` environment
variables, or by explicitly passing in the --editor program. The interactive IPython editor
requires IPython to be installed. You can also instead specify an input file with the
--infile / -i option.

Additional markup will be added that is appropriate to the choice of venue formatting. For
example, for the default 'gh' venue for GitHub Flavored Markdown, the final reprex output will
look like:

````
```python
arr = [1, 2, 3, 4, 5]
[x + 1 for x in arr]
#> [2, 3, 4, 5, 6]
max(arr) - min(arr)
#> 4
```

<sup>Created at 2021-02-27 00:13 PST by [reprexlite](https://github.com/jayqi/reprexlite) v{version}</sup>
````

The supported venue formats are:
{venue_formats}
"""  # noqa: E501


def get_help():
    help_text = HELP_TEMPLATE.format(
        version=get_version(),
        venue_formats="\n".join(
            f"- {key.value} : {entry.label}" for key, entry in formatter_registry.items()
        ),
    )
    return help_text


pyproject_toml_loader = cyclopts.config.Toml(
    "pyproject.toml",
    root_keys=("tool", "reprexlite"),
    search_parents=True,
    use_commands_as_keys=False,
)

reprexlite_toml_loader = cyclopts.config.Toml(
    "reprexlite.toml",
    search_parents=True,
    use_commands_as_keys=False,
)

dot_reprexlite_toml_loader = cyclopts.config.Toml(
    ".reprexlite.toml",
    search_parents=True,
    use_commands_as_keys=False,
)


user_reprexlite_toml_loader = cyclopts.config.Toml(
    Path(user_config_dir(appname="reprexlite")) / "config.toml",
    search_parents=False,
    use_commands_as_keys=False,
)

app = App(
    name="reprex",
    version=get_version,
    help_format="markdown",
    help=get_help(),
    config=(
        pyproject_toml_loader,
        reprexlite_toml_loader,
        dot_reprexlite_toml_loader,
        user_reprexlite_toml_loader,
    ),
)


def launch_ipython(config: ReprexConfig):
    try:
        from reprexlite.ipython import ReprexTerminalIPythonApp
    except IPythonNotFoundError:
        print(
            "IPythonNotFoundError: ipython is required to be installed to use the IPython "
            "interactive editor."
        )
        sys.exit(1)
    ReprexTerminalIPythonApp.set_reprex_config(config)
    ReprexTerminalIPythonApp.launch_instance(argv=[])
    sys.exit(0)


def launch_editor(editor) -> str:
    fw, name = tempfile.mkstemp(prefix="reprexlite-", suffix=".py")
    try:
        os.close(fw)  # Close the file descriptor
        # Open editor and edit the file
        proc = subprocess.Popen(args=f"{editor} {name}", shell=True)
        exit_code = proc.wait()
        if exit_code != 0:
            raise EditorError(f"{editor}: Editing failed with exit code {exit_code}")

        # Read the file back in
        with open(name, "rb") as fp:
            content = fp.read()
        return content.decode("utf-8-sig").replace("\r\n", "\n")
    except OSError as e:
        raise EditorError(f"{editor}: Editing failed: {e}") from e
    finally:
        os.unlink(name)


def get_editor() -> str:
    """Determine an editor to use for editing code."""
    for key in "VISUAL", "EDITOR":
        env_val = os.environ.get(key)
        if env_val:
            return env_val
    if platform.system() == "Windows":
        return "notepad"
    for editor in ("sensible-editor", "vim", "nano"):
        if os.system(f"which {editor} >/dev/null 2>&1") == 0:
            return editor
    return "vi"


def handle_editor(config: ReprexConfig) -> str:
    """Determines what to do based on the editor configuration."""
    editor = config.editor or get_editor()
    if editor == "ipython":
        launch_ipython(config)
        sys.exit(0)
    else:
        return launch_editor(editor)


@app.default
def main(
    *,
    infile: Annotated[
        Optional[Path],
        Parameter(
            name=("--infile", "-i"),
            help="Read code from this file instead of opening an editor.",
        ),
    ] = None,
    outfile: Annotated[
        Optional[Path],
        Parameter(
            name=("--outfile", "-o"),
            help="Write rendered reprex to this file instead of standard out.",
        ),
    ] = None,
    config: Annotated[ReprexConfig, Parameter(name="*")] = ReprexConfig(),
    verbose: Annotated[
        tuple[bool, ...],
        Parameter(
            name=("--verbose",), show_default=False, negative=False, help="Increase verbosity."
        ),
    ] = (),
    debug: Annotated[bool, Parameter(show=False)] = False,
):
    verbosity = sum(verbose)
    if verbosity:
        sys.stderr.write("infile: {}\n".format(infile))
        sys.stderr.write("outfile: {}\n".format(outfile))
        sys.stderr.write("config: {}\n".format(config))

    if debug:
        data = {"infile": infile, "outfile": outfile, "config": dataclasses.asdict(config)}
        sys.stdout.write(json.dumps(data))
        return data

    if infile:
        if verbose:
            sys.stderr.write(f"Reading from input file: {infile}")
        with infile.open("r") as fp:
            input = fp.read()
    else:
        input = handle_editor(config)
        if input.strip() == "":
            print("No input provided or saved via the editor. Exiting.")
            sys.exit(0)

    try:
        r = Reprex.from_input(input=input, config=config)
    except InputSyntaxError as e:
        print("ERROR: reprexlite has encountered an error while evaluating your input.")
        print(e)
        raise

    if outfile:
        with outfile.open("w") as fp:
            fp.write(r.render_and_format(terminal=False))
        print(f"Wrote rendered reprex to {outfile}")
    else:
        print(r.render_and_format(terminal=True), end="")

    return r


def entrypoint():
    """Entrypoint for the reprex command-line interface. This function is configured as the reprex
    entrypoint under [project.scripts].
    https://packaging.python.org/en/latest/specifications/entry-points/#use-for-scripts
    """
    app()



from enum import Enum
from typing import Iterator, Optional, Tuple

from reprexlite.exceptions import (
    InvalidInputPrefixesError,
    NoPrefixMatchError,
    PromptLengthMismatchError,
    UnexpectedError,
)


def removeprefix(s: str, prefix: str) -> str:
    """Utility function to strip a prefix from a string, whether or not there is a single
    whitespace character.
    """
    if s.startswith(prefix + " "):
        return s[len(prefix) + 1 :]
    elif s.startswith(prefix):
        return s[len(prefix) :]
    else:
        raise UnexpectedError(  # pragma: nocover
            "removeprefix should not be called on input that does not match the prefix. "
        )


class LineType(Enum):
    """An enum for different types of lines in text input to [parse][reprexlite.parsing.parse].

    Args:
        CODE (str): Line is code.
        RESULT (str): Line is the result of executing code.
    """

    CODE = "CODE"
    RESULT = "RESULT"


def parse(
    input: str,
    prompt: Optional[str],
    continuation: Optional[str],
    comment: Optional[str],
) -> Iterator[Tuple[str, LineType]]:
    """Generator function that parses input into lines of code or results.

    Args:
        input (str): String to parse
        prompt (Optional[str]): Prefix used as primary prompt of code lines
        continuation (Optional[str]): Prefix used as continuation prompt of code lines
        comment (Optional[str]): Prefix used to indicate result lines

    Yields:
        Iterator[Tuple[str, LineType]]: tuple of parsed line and line type
    """
    if not any((prompt, continuation, comment)):
        raise InvalidInputPrefixesError(
            "Cannot parse input if all of prompt, continuation, and comment are blank."
        )
    if len(prompt or "") != len(continuation or ""):
        raise PromptLengthMismatchError(
            f"Primary prompt ('{prompt}') and continuation prompt ('{continuation}') must be "
            "equal lengths."
        )

    for line_no, line in enumerate(input.split("\n")):
        # Case 1: With Prompt/Continuation, no Comment (e.g., doctest style)
        if prompt and continuation and not comment:
            if line.startswith(prompt):
                yield removeprefix(line, prompt), LineType.CODE
            elif line.startswith(continuation):
                yield removeprefix(line, continuation), LineType.CODE
            elif line == "":
                yield line, LineType.CODE
            else:
                yield line, LineType.RESULT

        # Case 2: No Prompt or Continuation, with Comment (e.g., reprex style)
        elif not prompt and not continuation and comment:
            if line.startswith(comment):
                yield removeprefix(line, comment), LineType.RESULT
            else:
                yield line, LineType.CODE

        # Case 3: Both Prompt/Contiuation and Comment
        elif prompt and continuation and comment:
            if line.startswith(prompt):
                yield removeprefix(line, prompt), LineType.CODE
            elif line.startswith(continuation):
                yield removeprefix(line, continuation), LineType.CODE
            elif line.startswith(comment):
                yield removeprefix(line, comment), LineType.RESULT
            elif line == "":
                yield line, LineType.CODE
            else:
                raise NoPrefixMatchError(
                    f"Line {line_no + 1} does not match any of prompt, continuation, or comment "
                    f"prefixes: '{line}'"
                )

        else:
            raise UnexpectedError("Unexpected case when using parse.")  # pragma: nocover


def parse_reprex(input: str) -> Iterator[Tuple[str, LineType]]:
    """Wrapper around [parse][reprexlite.parsing.parse] for parsing reprex-style input."""
    yield from parse(input=input, prompt=None, continuation=None, comment="#>")


def parse_doctest(input: str) -> Iterator[Tuple[str, LineType]]:
    """Wrapper around [parse][reprexlite.parsing.parse] for parsing doctest-style input."""
    yield from parse(input=input, prompt=">>>", continuation="...", comment=None)


def auto_parse(input: str) -> Iterator[Tuple[str, LineType]]:
    """Automatically parse input that is either doctest-style and reprex-style."""
    if any(line.startswith(">>>") for line in input.split("\n")):
        yield from parse_doctest(input)
    else:
        yield from parse_reprex(input)



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Optional, Protocol

from reprexlite.config import ReprexConfig, Venue
from reprexlite.exceptions import PygmentsNotFoundError
from reprexlite.session_info import SessionInfo
from reprexlite.version import __version__


class Formatter(Protocol):
    def __call__(self, reprex_str: str, config: Optional[ReprexConfig] = None) -> str: ...


@dataclass
class FormatterRegistration:
    fn: Formatter
    label: str


class FormatterRegistry:
    """Registry of formatters keyed by venue keywords."""

    _registry: Dict[str, FormatterRegistration] = {}

    def __getitem__(self, key: Venue) -> FormatterRegistration:
        return self._registry[Venue(key)]

    def __contains__(self, key: Venue) -> bool:
        return Venue(key) in self._registry

    def items(self):
        return self._registry.items()

    def keys(self):
        return self._registry.keys()

    def values(self):
        return self._registry.values()

    def register(self, venue: Venue, label: str):
        """Decorator that registers a formatter implementation.

        Args:
            venue (str): Venue keyword that formatter will be registered to.
            label (str): Short human-readable label explaining the venue.
        """

        def _register(fn: Formatter):
            self._registry[Venue(venue)] = FormatterRegistration(fn=fn, label=label)
            return fn

        return _register


formatter_registry = FormatterRegistry()


@formatter_registry.register(venue=Venue.DS, label=f"Discourse (alias for '{Venue.GH.value}')")
@formatter_registry.register(venue=Venue.SO, label=f"StackOverflow (alias for '{Venue.GH.value}')")
@formatter_registry.register(venue=Venue.GH, label="Github Flavored Markdown")
def format_as_markdown(
    reprex_str: str,
    config: Optional[ReprexConfig] = None,
) -> str:
    """
    Format a rendered reprex reprex as a GitHub-Flavored Markdown code block. By default, includes
    a footer that credits reprexlite.

    Args:
        reprex_str (str): The reprex string to render.
        config (Optional[ReprexConfig]): Configuration for the reprex. Defaults to None.

    Returns:
        str: The rendered reprex

    Example:
        ```python
        2+2
        #> 4
        ```
    """
    if config is None:
        config = ReprexConfig()
    advertise = config.advertise if config.advertise is not None else True
    out = []
    out.append("```python")
    out.append(reprex_str)
    out.append("```")
    if advertise:
        out.append("\n" + Advertisement().markdown())
    if config.session_info:
        out.append("\n<details><summary>Session Info</summary>")
        out.append("```text")
        out.append(str(SessionInfo()))
        out.append("```")
        out.append("</details>")
    return "\n".join(out) + "\n"


@formatter_registry.register(venue=Venue.HTML, label="HTML")
def format_as_html(reprex_str: str, config: Optional[ReprexConfig] = None) -> str:
    """Format a rendered reprex reprex as an HTML code block. If optional dependency Pygments is
    available, the rendered HTML will have syntax highlighting for the Python code. By default,
    includes a footer that credits reprexlite.

    Args:
        reprex_str (str): The reprex string to render.
        config (Optional[ReprexConfig]): Configuration for the reprex. Defaults to None.

    Returns:
        str: The rendered reprex

    Example:
        <pre><code>2+2
        #> 4</code></pre>
    """
    if config is None:
        config = ReprexConfig()
    advertise = config.advertise if config.advertise is not None else True
    out = []
    try:
        from pygments import highlight
        from pygments.formatters import HtmlFormatter
        from pygments.lexers import PythonLexer

        formatter = HtmlFormatter(style="friendly", lineanchors=True, linenos=True, wrapcode=True)
        out.append(f"<style>{formatter.get_style_defs('.highlight')}</style>")
        out.append(highlight(str(reprex_str), PythonLexer(), formatter))
    except ImportError:
        out.append(f"<pre><code>{reprex_str}</code></pre>")

    if advertise:
        out.append(Advertisement().html().strip())
    if config.session_info:
        out.append("<details><summary>Session Info</summary>")
        out.append(f"<pre><code>{SessionInfo()}</code></pre>")
        out.append("</details>")
    return "\n".join(out) + "\n"


@formatter_registry.register(venue=Venue.PY, label="Python script")
def format_as_python_script(reprex_str: str, config: Optional[ReprexConfig] = None) -> str:
    """Format a rendered reprex reprex as a Python script.

    Args:
        reprex_str (str): The reprex string to render.
        config (Optional[ReprexConfig]): Configuration for the reprex. Defaults to None.

    Returns:
        str: The rendered reprex

    Example:
        2+2
        #> 4
    """
    if config is None:
        config = ReprexConfig()
    advertise = config.advertise if config.advertise is not None else False
    out = [str(reprex_str)]
    if advertise:
        out.append("\n" + Advertisement().code_comment())
    if config.session_info:
        out.append("")
        sess_lines = str(SessionInfo()).split("\n")
        out.extend("# " + line for line in sess_lines)
    return "\n".join(out) + "\n"


@formatter_registry.register(venue=Venue.RTF, label="Rich Text Format")
def format_as_rtf(reprex_str: str, config: Optional[ReprexConfig] = None) -> str:
    """Format a rendered reprex reprex as a Rich Text Format (RTF) document. Requires dependency
    Pygments."""
    if config is None:
        config = ReprexConfig()
    advertise = config.advertise if config.advertise is not None else False
    try:
        from pygments import highlight
        from pygments.formatters import RtfFormatter
        from pygments.lexers import PythonLexer
    except ModuleNotFoundError as e:
        if e.name == "pygments":
            raise PygmentsNotFoundError("Pygments is required for RTF output.", name="pygments")
        else:
            raise

    out = str(reprex_str)
    if advertise:
        out += "\n\n" + Advertisement().text()
    if config.session_info:
        out += "\n\n" + str(SessionInfo())
    return highlight(out, PythonLexer(), RtfFormatter()) + "\n"


@formatter_registry.register(venue=Venue.SLACK, label="Slack")
def format_for_slack(reprex_str: str, config: Optional[ReprexConfig] = None) -> str:
    """Format a rendered reprex as Slack markup.

    Args:
        reprex_str (str): The reprex string to render.
        config (Optional[ReprexConfig]): Configuration for the reprex. Defaults to None.

    Returns:
        str: The rendered reprex

    Example:
        ```
        2+2
        #> 4
        ```
    """
    if config is None:
        config = ReprexConfig()
    advertise = config.advertise if config.advertise is not None else False
    out = []
    out.append("```")
    out.append(str(reprex_str))
    out.append("```")
    if advertise:
        out.append("\n" + Advertisement().text())
    if config.session_info:
        out.append("\n```")
        out.append(str(SessionInfo()))
        out.append("```")
    return "\n".join(out) + "\n"


class Advertisement:
    """Class for generating the advertisement note for reprexlite.

    Attributes:
        timestamp (str): Timestamp of instance instantiation
        version (str): Version of reprexlite
    """

    pkg = "reprexlite"
    url = "https://github.com/jayqi/reprexlite"

    def __init__(self):
        self.timestamp = datetime.now().astimezone().strftime("%Y-%m-%d %H:%M %Z")
        self.version = f"v{__version__}"

    def markdown(self) -> str:
        """Render reprexlite advertisement in GitHub Flavored Markdown."""
        return f"<sup>Created at {self.timestamp} by [{self.pkg}]({self.url}) {self.version}</sup>"

    def html(self) -> str:
        """Render reprexlite advertisement in HTML."""
        return (
            f"<p><sup>Created at {self.timestamp} by "
            f'<a href="{self.url}">{self.pkg}</a> {self.version}</sup></p>'
        )

    def code_comment(self) -> str:
        """Render reprexlite advertisement as a comment in Python code."""
        return f"# {self.text()}"

    def text(self) -> str:
        """Render reprexlite advertisement in plain text."""
        return f"Created at {self.timestamp} by {self.pkg} {self.version} <{self.url}>"



from dataclasses import dataclass
from enum import Enum
from typing import Annotated, Optional

from cyclopts import Parameter

from reprexlite.exceptions import (
    InvalidParsingMethodError,
    InvalidVenueError,
    PromptLengthMismatchError,
)


class ParsingMethod(str, Enum):
    """Methods for parsing input strings.

    Args:
        AUTO (str): Automatically identify reprex-style or doctest-style input.
        DECLARED (str): Use configured values for parsing.
    """

    AUTO = "auto"
    DECLARED = "declared"


class Venue(str, Enum):
    """Enum for specifying the output venue for a reprex."""

    GH = "gh"
    DS = "ds"
    SO = "so"
    HTML = "html"
    PY = "py"
    RTF = "rtf"
    SLACK = "slack"


@dataclass
class ReprexConfig:
    """Configuration dataclass for reprexlite. Used to configure input parsing and output
    formatting.

    Args:
        editor (str): Command-line program name of editor to use. If not specified, check $EDITOR
            and $VISUAL environment variables. If 'ipython', will launch the IPython interactive
            editor.
        venue (str): Key to identify the output venue that the reprex will be shared in. Used to
            select an appropriate formatter. See "Venues Formatting" documentation for formats
            included with reprexlite.
        advertise (bool): Whether to include a footer that credits reprexlite. If unspecified, will
            depend on specified venue formatter's default.
        session_info (bool): Include details about session and environment that the reprex was
            generated with.
        style (bool): Whether to autoformat code with black. Requires black to be installed.
        prompt (str): Prefix to use as primary prompt for code lines.
        continuation (str): Prefix to use as secondary prompt for continued code lines.
        comment (str): Prefix to use for results returned by expressions.
        keep_old_results (bool): Whether to additionally include results of expressions detected in
            the original input when formatting the reprex output.
        parsing_method (str): Method for parsing input. 'auto' will automatically detect either
            default reprex-style input or standard doctest-style input. 'declared' will allow you
            to specify custom line prefixes. Values for 'prompt', 'continuation', and 'comment'
            will be used for both output formatting and input parsing, unless the associated
            'input_*' override settings are supplied.
        input_prompt (str): Prefix to use as primary prompt for code lines when parsing input. Only
            used if 'parsing_method' is 'declared'. If not set, 'prompt' is used for both input
            parsing and output formatting.
        input_continuation (str): Prefix to use as secondary prompt for continued code lines when
            parsing input. Only used if 'parsing_method' is 'declared'. If not set, 'prompt' is
            used for both input parsing and output formatting.
        input_comment (str): Prefix to use for results returned by expressions when parsing input.
            Only used if 'parsing_method' is 'declared'. If not set, 'prompt' is used for both
            input parsing and output formatting.
    """

    editor: Annotated[Optional[str], Parameter(name=("--editor", "-e"))] = None
    # Formatting
    venue: Annotated[Venue, Parameter(name=("--venue", "-v"))] = Venue.GH
    advertise: Optional[bool] = None
    session_info: bool = False
    style: bool = False
    prompt: str = ""
    continuation: str = ""
    comment: str = "#>"
    keep_old_results: bool = False
    # Parsing
    parsing_method: ParsingMethod = ParsingMethod.AUTO
    input_prompt: Optional[str] = None
    input_continuation: Optional[str] = None
    input_comment: Optional[str] = None

    def __post_init__(self):
        # Validate venue
        try:
            Venue(self.venue)
        except ValueError:
            raise InvalidVenueError(
                f"{self.venue} is not a valid value for parsing method."
                f"Valid values are: {list(m.value for m in Venue)}"
            )
        # Validate prompt and continuation prefixes
        if len(self.prompt) != len(self.continuation):
            raise PromptLengthMismatchError(
                f"Primary prompt ('{self.prompt}') and continuation prompt "
                f"('{self.continuation}') must be equal lengths."
            )
        # Validate parsing method
        try:
            ParsingMethod(self.parsing_method)
        except ValueError:
            raise InvalidParsingMethodError(
                f"{self.parsing_method} is not a valid value for parsing method."
                f"Valid values are: {[pm.value for pm in ParsingMethod]}"
            )

    @property
    def resolved_input_prompt(self):
        if self.input_prompt is not None:
            return self.input_prompt
        return self.prompt

    @property
    def resolved_input_continuation(self):
        if self.input_continuation is not None:
            return self.input_continuation
        return self.continuation

    @property
    def resolved_input_comment(self):
        if self.input_comment is not None:
            return self.input_comment
        return self.comment



from contextlib import contextmanager, redirect_stdout
import io
import re
from typing import Optional

import reprexlite.cli
from reprexlite.config import ReprexConfig
from reprexlite.exceptions import IPythonNotFoundError
from reprexlite.reprexes import Reprex
from reprexlite.version import __version__

try:
    from IPython import InteractiveShell
    from IPython.core.magic import Magics, line_cell_magic, magics_class
    from IPython.core.release import version as ipython_version
    from IPython.core.usage import default_banner_parts
    from IPython.terminal.interactiveshell import TerminalInteractiveShell
    from IPython.terminal.ipapp import TerminalIPythonApp
except ModuleNotFoundError as e:
    if e.name == "IPython":
        raise IPythonNotFoundError(*e.args, name="IPython")
    else:
        raise


@contextmanager
def patch_edit(input: str):
    """Patches typer.edit to return the input string instead of opening up the text editor. This
    is a trick to hook up the IPython cell magic's cell contents to the typer CLI app.
    """

    def return_input(*args, **kwargs) -> str:
        return input

    original = reprexlite.cli.handle_editor
    setattr(reprexlite.cli, "handle_editor", return_input)
    yield
    setattr(reprexlite.cli, "handle_editor", original)


@magics_class
class ReprexMagics(Magics):
    @line_cell_magic
    def reprex(self, line: str, cell: Optional[str] = None):
        """reprex IPython magic. Use line magic %reprex to print help. Use cell magic %%reprex to
        render a reprex."""
        # Line magic, print help
        if cell is None:
            with io.StringIO() as buffer, redirect_stdout(buffer):
                reprexlite.cli.app("--help")
                help_text = buffer.getvalue()
            help_text = re.sub(r"^Usage: reprex", r"Cell Magic Usage: %%reprex", help_text)
            print(f"reprexlite v{__version__} IPython Magic\n\n" + help_text)
            return
        # Cell magic, render reprex
        with patch_edit(cell):
            reprexlite.cli.app(line.split())
            # print(stdout, end="")


def load_ipython_extension(ipython: InteractiveShell):
    """Special function to register this module as an IPython extension.
    https://ipython.readthedocs.io/en/stable/config/extensions/#writing-extensions
    """

    ipython.register_magics(ReprexMagics)


ipython_banner_parts = [
    default_banner_parts[0],
    f"reprexlite {__version__} -- Interactive reprex editor via IPython {ipython_version}.",
]


class ReprexTerminalInteractiveShell(TerminalInteractiveShell):
    """Subclass of IPython's TerminalInteractiveShell that automatically executes all cells using
    reprexlite instead of normally."""

    banner1 = "".join(ipython_banner_parts)  # type: ignore
    _reprex_config: Optional[ReprexConfig] = None

    def run_cell(self, raw_cell: str, *args, **kwargs):
        # "exit()" and Ctrl+D short-circuit this and don't need to be handled
        if raw_cell != "exit":
            try:
                r = Reprex.from_input(raw_cell, config=self.reprex_config)
                print(r.render_and_format(terminal=True), end="")
            except Exception as e:
                print("ERROR: reprexlite has encountered an error while evaluating your input.")
                print(e, end="")

            # Store history
            self.history_manager.store_inputs(self.execution_count, raw_cell, raw_cell)  # type: ignore
            self.execution_count += 1

            return None
        else:
            return super().run_cell(raw_cell, *args, **kwargs)

    @property
    def reprex_config(self) -> ReprexConfig:
        if self._reprex_config is None:
            self._reprex_config = ReprexConfig()
        return self._reprex_config


class ReprexTerminalIPythonApp(TerminalIPythonApp):
    """Subclass of TerminalIPythonApp that launches ReprexTerminalInteractiveShell."""

    interactive_shell_class = ReprexTerminalInteractiveShell  # type: ignore

    @classmethod
    def set_reprex_config(cls, config: ReprexConfig):
        """Set the reprex config bound on the interactive shell."""
        cls.interactive_shell_class._reprex_config = config  # type: ignore
